import copy
import json
import math
import pickle
import numpy as np
import pandas as pd
from torch.utils.data import Dataset, DataLoader
import torch
import os
from collections import defaultdict as ddict
from tqdm import tqdm
import random


class UKGData(object):
    """Data preprocessing of ukg data.

    Attributes:
        args: Some pre-set parameters, such as dataset path, etc. 
        ent2id: Encoding the entity in triples, type: dict.
        rel2id: Encoding the relation in triples, type: dict.
        id2ent: Decoding the entity in triples, type: dict.
        id2rel: Decoding the realtion in triples, type: dict.
        train_triples: Record the triples for training, type: list.
        valid_triples: Record the triples for validation, type: list.
        test_triples: Record the triples for testing, type: list.
        PSL_triples: Record the triples for softlogic, type: list. (will be used in UKGE_PSL)
        pseudo_triples: Record the triples for pseudo, type: list. (will be used in UPGAT)
        all_true_triples: Record all triples including train,valid and test, type: list.
        hr2t_train: Record the tail corresponding to the same head and relation, type: defaultdict(class:set).
        rt2h_train: Record the head corresponding to the same tail and relation, type: defaultdict(class:set).
        h2rt_train: Record the tail, relation corresponding to the same head, type: defaultdict(class:set).
        t2rh_train: Record the head, realtion corresponding to the same tail, type: defaultdict(class:set).
        hr2t_total: Record the tail corresponding to the same head and relation in whole dataset(train + val + test), type: defaultdict(class:set).
        rt2h_total: Record the head corresponding to the same tail and relation in whole dataset(train + val + test), type: defaultdict(class:set).
        RatioOfPSL: Record the ratio of the number of PSL samples to the number of training samples. (will be used in UKGE_PSL)
        pseudo_dataiter: Record the data in dataloader. (will be used in UPGAT)
    """

    def __init__(self, args):
        self.args = args

        # basic part
        self.ent2id = {}
        self.rel2id = {}
        self.id2ent = {}
        self.id2rel = {}
        # store the ID of the sample
        self.train_triples = []
        self.valid_triples = []
        self.test_triples = []
        self.PSL_triples = []
        self.all_true_triples = set()
        # for PSL
        self.RatioOfPSL = 0
        # for UPGAT
        self.pseudo_triples = []
        self.pseudo_dataiter = None
        # for calculating nDCG
        # self.hr_map = {}
        self.hr2t_train = ddict(set)
        self.rt2h_train = ddict(set)
        self.h2rt_train = ddict(set)
        self.t2rh_train = ddict(set)
        self.hr2t_total = ddict(set)
        self.rt2h_total = ddict(set)

        self.get_id()
        # self.get_ndcg_test_data()
        if args.use_weight:
            self.count = self.count_frequency(self.train_triples)

    def get_id(self):
        """Get entity/relation id, and entity/relation number.

        Update:
            self.ent2id: Entity to id.
            self.rel2id: Relation to id.
            self.id2ent: id to Entity.
            self.id2rel: id to Relation.
            self.args.num_ent: Entity number.
            self.args.num_rel: Relation number.
        """
        last_e = -1
        last_r = -1

        with open(os.path.join(self.args.data_path, "train.tsv"), encoding='utf-8') as fin:
            for line in fin:
                line = line.rstrip('\n').split('\t')
                h, r, t, w = line[0], line[1], line[2], float(line[3])

                if self.ent2id.get(h) == None:
                    last_e += 1
                    self.ent2id[h] = last_e
                    self.id2ent[last_e] = h

                if self.ent2id.get(t) == None:
                    last_e += 1
                    self.ent2id[t] = last_e
                    self.id2ent[last_e] = t

                if self.rel2id.get(r) == None:
                    last_r += 1
                    self.rel2id[r] = last_r
                    self.id2rel[last_r] = r

                h = self.ent2id[h]
                r = self.rel2id[r]
                t = self.ent2id[t]

                self.train_triples.append((h, r, t, w))

        with open(os.path.join(self.args.data_path, "val.tsv"), encoding='utf-8') as fin:
            for line in fin:
                line = line.rstrip('\n').split('\t')
                h, r, t, w = line[0], line[1], line[2], float(line[3])

                if self.ent2id.get(h) == None:
                    last_e += 1
                    self.ent2id[h] = last_e
                    self.id2ent[last_e] = h

                if self.ent2id.get(t) == None:
                    last_e += 1
                    self.ent2id[t] = last_e
                    self.id2ent[last_e] = t

                if self.rel2id.get(r) == None:
                    last_r += 1
                    self.rel2id[r] = last_r
                    self.id2rel[last_r] = r

                h = self.ent2id[h]
                r = self.rel2id[r]
                t = self.ent2id[t]

                self.valid_triples.append((h, r, t, w))

        with open(os.path.join(self.args.data_path, "test.tsv"), encoding='utf-8') as fin:
            for line in fin:
                line = line.rstrip('\n').split('\t')
                h, r, t, w = line[0], line[1], line[2], float(line[3])

                if self.ent2id.get(h) == None:
                    last_e += 1
                    self.ent2id[h] = last_e
                    self.id2ent[last_e] = h

                if self.ent2id.get(t) == None:
                    last_e += 1
                    self.ent2id[t] = last_e
                    self.id2ent[last_e] = t

                if self.rel2id.get(r) == None:
                    last_r += 1
                    self.rel2id[r] = last_r
                    self.id2rel[last_r] = r

                h = self.ent2id[h]
                r = self.rel2id[r]
                t = self.ent2id[t]

                self.test_triples.append((h, r, t, w))

        softlogic_temp_path = os.path.join(self.args.data_path, "softlogic.tsv")
        if os.path.exists(softlogic_temp_path):
            with open(os.path.join(self.args.data_path, "softlogic.tsv"), encoding='utf-8') as fin:
                for line in fin:
                    line = line.rstrip('\n').split('\t')
                    h, r, t, w = line[0], line[1], line[2], float(line[3])

                    if self.ent2id.get(h) == None:
                        last_e += 1
                        self.ent2id[h] = last_e
                        self.id2ent[last_e] = h

                    if self.ent2id.get(t) == None:
                        last_e += 1
                        self.ent2id[t] = last_e
                        self.id2ent[last_e] = t

                    if self.rel2id.get(r) == None:
                        last_r += 1
                        self.rel2id[r] = last_r
                        self.id2rel = r

                    h = self.ent2id[h]
                    r = self.rel2id[r]
                    t = self.ent2id[t]
                    self.PSL_triples.append((h, r, t, w))

        self.RatioOfPSL = len(self.PSL_triples) / len(self.train_triples) # calculate the ratio of the number of PSL samples to the number of training samples.

        # When train student_model in UPGAT, get pseudo_triples as below.
        if self.args.model_name == 'UPGAT' and not self.args.teacher_model:
            with open(os.path.join(self.args.data_path, "pseudo.tsv"), encoding='utf-8') as fin:
                for line in fin:
                    line = line.rstrip('\n').split('\t')
                    h, r, t, w = int(line[0]), int(line[1]), int(line[2]), float(line[3])
                    self.pseudo_triples.append((h, r, t, w))
            self.args.pseudo_bs = len(self.pseudo_triples) // (len(self.train_triples)//self.args.train_bs)
            pseudo_dataloader = DataLoader(self.pseudo_triples, batch_size=self.args.pseudo_bs, num_workers=self.args.num_workers, pin_memory=True, shuffle=True, drop_last=True)
            self.pseudo_dataiter = iter(pseudo_dataloader)
            self.all_true_triples = set(
                self.train_triples + self.valid_triples + self.test_triples + self.pseudo_triples
            )
        else:
            self.all_true_triples = set(
                self.train_triples + self.valid_triples + self.test_triples
            )

        self.args.num_ent = len(self.ent2id)
        self.args.num_rel = len(self.rel2id)

    # def get_ndcg_test_data(self):
    #     # 使用test_triples构造成hr_map
    #     with open(os.path.join(self.args.data_path, "ndcg_test.pickle"), 'rb') as f:
    #         self.hr_map = pickle.load(f)  # unpickle

        # self.hr_map = {}
        # for triple in self.test_triples:
        #     h, r, t, w = triple
        #     if self.hr_map.get(h) is None:
        #         self.hr_map[h] = {}
        #     if self.hr_map[h].get(r) is None:
        #         self.hr_map[h][r] = {t: w}
        #     else:
        #         self.hr_map[h][r][t] = w

        # tmp = 0
        # for h in self.hr_map:
        #     for r in self.hr_map[h]:
        #         tmp = max(tmp, len(self.hr_map[h][r].keys()))
        #
        # print(self.hr_map)
        # print(len(self.hr_map.keys()))
        # print(tmp)
        # exit(0)


    def get_hr2t_rt2h_from_train(self):
        """Get the set of hr2t and rt2h from train dataset, the data type is numpy.

        Update:
            self.hr2t_train: The set of hr2t.
            self.rt2h_train: The set of rt2h.
        """

        for h, r, t, w in self.train_triples:
            self.hr2t_train[(h, r)].add(t)
            self.rt2h_train[(r, t)].add(h)
        for h, r in self.hr2t_train:
            self.hr2t_train[(h, r)] = np.array(list(self.hr2t_train[(h, r)]))
        for r, t in self.rt2h_train:
            self.rt2h_train[(r, t)] = np.array(list(self.rt2h_train[(r, t)]))

    # Add by Messi
    def get_hr2t_rt2h_from_total(self):
        """Get the set of hr2t and rt2h from whole dataset(train+val+test), the data type is numpy.

        Update:
            self.hr2t_total: The set of hr2t in whole dataset.
            self.rt2h_total: The set of rt2h in whole dataset.
        """
        for h, r, t, w in self.all_true_triples:
            self.hr2t_total[(h, r)].add(t)
            self.rt2h_total[(r, t)].add(h)
        for h, r in self.hr2t_total:
            self.hr2t_total[(h, r)] = np.array(list(self.hr2t_total[(h, r)]))
        for r, t in self.rt2h_total:
            self.rt2h_total[(r, t)] = np.array(list(self.rt2h_total[(r, t)]))

    @staticmethod
    def count_frequency(triples, start=4):
        '''Get frequency of a partial triple like (head, relation) or (relation, tail).

        The frequency will be used for subsampling like word2vec.

        Args:
            triples: Sampled triples.
            start: Initial count number.

        Returns:
            count: Record the number of (head, relation).
        '''
        count = {}
        for head, relation, tail, w in triples:
            if (head, relation) not in count:
                count[(head, relation)] = start
            else:
                count[(head, relation)] += 1

            if (tail, -relation - 1) not in count:
                count[(tail, -relation - 1)] = start
            else:
                count[(tail, -relation - 1)] += 1
        return count

    def get_h2rt_t2hr_from_train(self):
        """Get the set of h2rt and t2hr from train dataset, the data type is numpy.

        Update:
            self.h2rt_train: The set of h2rt.
            self.t2rh_train: The set of t2hr.
        """
        for h, r, t, w in self.train_triples:
            self.h2rt_train[h].add((r, t))
            self.t2rh_train[t].add((r, h))
        for h in self.h2rt_train:
            self.h2rt_train[h] = np.array(list(self.h2rt_train[h]))
        for t in self.t2rh_train:
            self.t2rh_train[t] = np.array(list(self.t2rh_train[t]))

    def get_hr_train(self):
        '''Change the generation mode of batch.
        Merging triples which have same head and relation for 1vsN training mode.

        update:
            self.train_triples: The tuple(hr, t) list for training
        '''
        self.t_triples = self.train_triples
        self.train_triples = [(hr, list(t)) for (hr, t) in self.hr2t_train.items()]


class UKGEBaseSampler(UKGData):
    """
        data processing for UKG data, the sampling method is consistent with NeuralKG (https://github.com/zjukg/NeuralKG)
    """
    def __init__(self, args):
        super().__init__(args)
        self.get_hr2t_rt2h_from_train()
        self.get_hr2t_rt2h_from_total()

    def corrupt_head(self, t, r, num_max=1):
        """Negative sampling of head entities.

        Args:
            t: Tail entity in triple.
            r: Relation in triple.
            num_max: The maximum of negative samples generated

        Returns:
            neg: The negative sample of head entity filtering out the positive head entity.
        """
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.rt2h_total[(r, t)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def corrupt_tail(self, h, r, num_max=1):
        """Negative sampling of tail entities.

        Args:
            h: Head entity in triple.
            r: Relation in triple.
            num_max: The maximum of negative samples generated

        Returns:
            neg: The negative sample of tail entity filtering out the positive tail entity.
        """
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.hr2t_total[(h, r)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def head_batch(self, h, r, t, neg_size=None):
        """Negative sampling of head entities.

        Args:
            h: Head entity in triple
            t: Tail entity in triple.
            r: Relation in triple.
            neg_size: The size of negative samples.

        Returns:
            The negative sample of head entity. [neg_size]
        """
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_head(t, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def tail_batch(self, h, r, t, neg_size=None):
        """Negative sampling of tail entities.

        Args:
            h: Head entity in triple
            t: Tail entity in triple.
            r: Relation in triple.
            neg_size: The size of negative samples.

        Returns:
            The negative sample of tail entity. [neg_size]
        """
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_tail(h, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def get_train(self):
        return self.train_triples

    def get_valid(self):
        return self.valid_triples

    def get_test(self):
        return self.test_triples

    def get_PSL(self):
        return self.PSL_triples

    def get_all_true_triples(self):
        return self.all_true_triples

    def get_hr_map(self):
        return self.hr_map


class GMUCData(object):
    """Data processing for few-shot GMUC & GMUC+ data.

    Attributes:
        args: Some pre-set parameters, such as dataset path, etc.
        ent2id: Encoding the entity in triples, type: dict.
        rel2id: Encoding the relation in triples, type: dict.
        symbol2id: Encoding the entity and relation in triples, type: dict.
        e1rel_e2: Record the tail corresponding to the same head and relation, type: defaultdict(class:list).
        rele1_e2: Record the tail corresponding to the same relation and head, type: defaultdict(class:dict).
        train_tasks: Record the triples for training, type: dict.
        dev_tasks: Record the triples for validation, type: dict.
        test_tasks: Record the triples for testing, type: dict.
        task_pool: A task is a relation, type: list.
        path_graph: Background triples, type: list.
        type2ents: All entities of the same type, type: defaultdict(class:set).
        known_rels: The triples of path_graph and train_tasks, type: defaultdict(class:list).
        num_tasks: The number of tasks, type: int.
        rel2candidates: Record the entities corresponding to the same relation, type: list.
        ent2ic: Calculate IIC for every entity, type: dict.
        rel_uc1: Calculate IIC for every relation, type: dict.
        rel_uc2: Calculate IIC for every relation, type: dict.
        connections: Neighbor information for each entity, type: numpy.
        e1_rele2: Record the relation and tail corresponding to the same head, type: defaultdict(class:list).
        e1_degrees: Record the number of neighbors per entity, type: defaultdict(class:int).
    """

    def __init__(self, args):

        self.args = args

        self.ent2id = {}
        self.rel2id = {}
        self.symbol2id = {}  # ent + rel
        self.rel2candidates = {}
        self.e1rel_e2 = ddict(list)
        self.rele1_e2 = ddict(dict)

        self.train_tasks = json.load(open(self.args.data_path + '/train_tasks.json'))
        self.dev_tasks = json.load(open(self.args.data_path + '/dev_tasks.json'))
        self.test_tasks = json.load(open(self.args.data_path + '/test_tasks.json'))
        self.task_pool = list(self.train_tasks.keys())
        self.num_tasks = len(self.task_pool)
        self.path_graph = open(self.args.data_path + '/path_graph').readlines()
        self.type2ents = ddict(set)
        self.known_rels = ddict(list)

        self.get_tasks()
        self.get_rel2candidates()
        self.get_e1rel_e2()
        self.get_rele1_e2()
        degrees = self.build_graph(max_=self.args.max_neighbor)

        if args.has_ont:  # for GMUC+
            self.ent2ic = {}
            self.rel_uc1 = {}
            self.rel_uc2 = {}
            self.get_ontology()

    def get_tasks(self):
        """Get entity/relation id, and entity/relation number.

        Update:
            self.ent2id: Entity to id.
            self.rel2id: Relation to id.
            self.symbol2id: Entity and relation to id.
            self.args.num_ent: Entity number.
            self.args.num_rel: Relation number.
        """

        eid = -1
        rid = -1

        for k, v in self.train_tasks.items():
            if self.rel2id.get(k) is None:
                rid += 1
                self.rel2id[k] = rid
            for triple in v:
                e1, rel, e2, s = triple[0], triple[1], triple[2], triple[3]
                if self.ent2id.get(e1) is None:
                    eid += 1
                    self.ent2id[e1] = eid

                if self.ent2id.get(e2) is None:
                    eid += 1
                    self.ent2id[e2] = eid

        for k, v in self.dev_tasks.items():
            if self.rel2id.get(k) is None:
                rid += 1
                self.rel2id[k] = rid
            for triple in v:
                e1, rel, e2, s = triple[0], triple[1], triple[2], triple[3]
                if self.ent2id.get(e1) is None:
                    eid += 1
                    self.ent2id[e1] = eid

                if self.ent2id.get(e2) is None:
                    eid += 1
                    self.ent2id[e2] = eid

        for k, v in self.test_tasks.items():
            if self.rel2id.get(k) is None:
                rid += 1
                self.rel2id[k] = rid
            for triple in v:
                e1, rel, e2, s = triple[0], triple[1], triple[2], triple[3]
                if self.ent2id.get(e1) is None:
                    eid += 1
                    self.ent2id[e1] = eid

                if self.ent2id.get(e2) is None:
                    eid += 1
                    self.ent2id[e2] = eid

        for line in self.path_graph:
            e1, k, e2, s = line.strip().split()
            if self.rel2id.get(k) is None:
                rid += 1
                self.rel2id[k] = rid
            if self.rel2id.get(k + '_inv') is None:
                rid += 1
                self.rel2id[k + '_inv'] = rid
            if self.ent2id.get(e1) is None:
                eid += 1
                self.ent2id[e1] = eid

            if self.ent2id.get(e2) is None:
                eid += 1
                self.ent2id[e2] = eid

        self.args.num_ent = len(self.ent2id)
        self.args.num_rel = len(self.rel2id)

        symbol_id = {}
        i = 0
        for key in self.rel2id.keys():
            if key not in ['', 'OOV']:
                symbol_id[key] = i
                i += 1

        for key in self.ent2id.keys():
            if key not in ['', 'OOV']:
                symbol_id[key] = i
                i += 1

        symbol_id['PAD'] = i
        self.symbol2id = symbol_id
        self.symbol2vec = None

        self.num_symbols = len(self.symbol2id.keys()) - 1  # one for 'PAD'
        self.args.num_symbols = self.num_symbols
        self.pad_id = self.num_symbols

    def get_rel2candidates(self):
        """Get the set of rel2candidates. A candidate is an entity under a relationship.

        Update:
            self.known_rels: The set of known_rels.
            self.type2ents: The set of type2ents.
            self.rel2candidates: Obtain 1000 entities of the same type as a candidate set.
        """
        # known_rels = path_graph + train_tasks
        for line in self.path_graph:
            e1, rel, e2, s = line.strip().split()
            self.known_rels[rel].append([e1, rel, e2, s])

        for key, triples in self.train_tasks.items():
            self.known_rels[key] = triples

        all_reason_relations = list(self.known_rels.keys()) + list(self.dev_tasks.keys()) + list(self.test_tasks.keys())
        all_reason_relation_triples = list(self.known_rels.values()) + list(self.dev_tasks.values()) + list(
            self.test_tasks.values())
        assert len(all_reason_relations) == len(all_reason_relation_triples)

        # rel2candidates = {rel: [tail entity candidates]}
        if self.args.dataset_name == "nl27k":
            ents = self.ent2id.keys()
            for ent in ents:
                type_ = ent.split(':')[1]
                self.type2ents[type_].add(ent)

            for rel, triples in zip(all_reason_relations, all_reason_relation_triples):
                # all possible tail entity type
                possible_types = set()
                for triple in triples:
                    tail_type = triple[2].split(':')[1]
                    possible_types.add(tail_type)

                # all possible tail entity candiates
                candidates = []
                for tail_type in possible_types:
                    candidates.extend(self.type2ents[tail_type])

                candidates = list(set(candidates))
                if len(candidates) > 1000:
                    candidates = candidates[:1000]
                self.rel2candidates[rel] = candidates
        else:
            for rel, triples in zip(all_reason_relations, all_reason_relation_triples):
                candidates = []

                for triple in triples:
                    h, r, t, s = triple
                    candidates.append(h)
                    candidates.append(t)

                candidates = list(set(candidates))
                if len(candidates) > 1000:
                    candidates = candidates[:1000]
                self.rel2candidates[rel] = candidates

    def get_e1rel_e2(self):
        """Get the set of e1rel_e2 from all dataset.

        Update:
            self.e1rel_e2: The set of e1rel_e2.
        """
        task_triples = list(self.train_tasks.values()) + list(self.dev_tasks.values()) + list(self.test_tasks.values())
        for task in task_triples:
            for triple in task:
                e1, rel, e2, s = triple
                self.e1rel_e2[e1 + rel].append(e2)

    def get_rele1_e2(self):
        """Get the set of get_rele1_e2 from dev and test dataset.

        Update:
            self.rele1_e2: The set of rele1_e2.
        """
        tasks = list(self.dev_tasks.values()) + list(self.test_tasks.values())

        for task in tasks:
            d = ddict(list)
            for triple in task:
                h, r, t, s = triple
                d[h].append([t, s])
            self.rele1_e2[r] = d

    def get_ontology(self):
        """Get the IIC of the entity and UC of the relation.

        IIC(c) = 1-\frac{log(hypo(c)+1)}{log(n)}

        UC_r1(r)=\sum_{h \in D_r, t \in R_r} (UC_e(h) + UC_e(t))
        UC_r2(r) = |D_r| * |R_r|

        Update:
            self.ent2ic: The set of rele1_e2.
            self.rel_uc1: The set of rele1_e2.
            self.rel_uc2: The set of rele1_e2.
        """
        df_ont = pd.read_csv(self.args.data_path + '/ontology.csv')
        pairs = []
        concept_set = set()

        for i in range(len(df_ont)):
            if df_ont.at[i, 'rel'] == 'is_A':
                c1 = df_ont.at[i, 'h']
                c2 = df_ont.at[i, 't']
                pairs.append((c1, c2))  # c1是c2的子类
                concept_set.add(c1)
                concept_set.add(c2)

        # Extract the range and domain of the relationship
        # Get all the relationships
        relation_set = set()
        df_domain = df_ont[df_ont.rel == 'domain']
        for i in df_domain.index:
            e = df_domain.at[i, 'h']
            relation_set.add(e)
        # Extract entity types
        entity_set = concept_set.difference(relation_set)

        # Calculate the sub-words for each entity type
        hypo_dict = dict()  # The next subordinate word of the entity
        for i in entity_set:
            hypo_list = []
            # Look for the subterfuge of this entity
            for pair in pairs:
                if pair[1] == i:
                    hypo_list.append(pair[0])
            hypo_dict[i] = hypo_list

        real_hypo_dict = copy.deepcopy(hypo_dict)
        for i in entity_set:
            for j in real_hypo_dict[i]:
                real_hypo_dict[i].extend(real_hypo_dict[j])
            real_hypo_dict[i] = list(set(real_hypo_dict[i]))

        # Calculate the IC value of the entity
        ent_type_ic = dict()
        for i in entity_set:
            ent_type_ic[i] = 1 - math.log(len(real_hypo_dict[i]) + 1) / math.log(292)

        ent2uc = ddict(float)
        rel2uc1 = ddict(float)  # uc(h)+uc(t)
        rel2uc2 = ddict(float)  # |domain|*|range|

        ent2type = {}
        df_type = df_ont[df_ont.rel == 'type']
        for i in df_type.index:
            h = df_type.at[i, 'h']
            t = df_type.at[i, 't']
            ent2type[h] = t

        ent = self.ent2id.keys()
        for i in ent:
            type_ = ent2type[i]
            ent2uc[i] = 1 - ent_type_ic[type_]
        ent_uc = list(ent2uc.values())
        mu = np.mean(ent_uc)
        sigma = np.std(ent_uc)
        for k, v in ent2uc.items():
            ent2uc[k] = (v - mu) / sigma
        self.ent2ic = ent2uc

        # Calculate the UC value of the relation
        for k, v in self.known_rels.items():
            rel = k
            domain_ = set()
            range_ = set()
            for _ in v:
                h, r, t, s = _
                h_type = h.split(':')[1]
                t_type = t.split(':')[1]
                domain_.add(h_type)
                range_.add(t_type)
            ic_sum = 0
            count = 0
            for i in domain_:
                for j in range_:
                    count += 1
                    ic_sum += (1 - ent_type_ic['concept:' + i] + 1 - ent_type_ic['concept:' + j]) / 2

            rel2uc1[rel] = ic_sum / count
            rel2uc2[rel] = count

        # normalization
        rel_uc1 = np.array(list(rel2uc1.values()))
        rel_uc2 = np.array(list(rel2uc2.values()))
        mu1, mu2 = np.mean(rel_uc1), np.mean(rel_uc2)
        sigma1, sigma2 = np.std(rel_uc1), np.std(rel_uc2)
        for k, v in rel2uc1.items():
            rel2uc1[k] = (v - mu1) / sigma1
        for k, v in rel2uc2.items():
            rel2uc2[k] = (v - mu2) / sigma2

        self.rel_uc1 = rel2uc1
        self.rel_uc2 = rel2uc2

    def build_graph(self, max_=50):
        """Build the graph according to path_graph.

        Update:
            self.connections: The set of connections.
            self.e1_rele2: The set of e1_rele2.
            self.e1_degrees: The set of e1_degrees.
        """

        self.connections = (np.ones((self.args.num_ent, max_, 3)) * self.pad_id)
        self.e1_rele2 = ddict(list)
        self.e1_degrees = ddict(int)

        with open(self.args.data_path + '/path_graph') as f:
            lines = f.readlines()
            for line in tqdm(lines):
                e1, rel, e2, s = line.strip().split()
                self.e1_rele2[e1].append((self.symbol2id[rel], self.symbol2id[e2], float(s)))
                self.e1_rele2[e2].append((self.symbol2id[rel + '_inv'], self.symbol2id[e1], float(s)))

        degrees = {}
        for ent, id_ in self.ent2id.items():
            # Take out a neighbor of a head entity
            neighbors = self.e1_rele2[ent]
            # The number of this neighbor is limited only to max_
            if len(neighbors) > max_:
                neighbors = neighbors[:max_]
            degrees[ent] = len(neighbors)
            self.e1_degrees[id_] = len(neighbors)
            for idx, _ in enumerate(neighbors):
                self.connections[id_, idx, 0] = _[0]  # relation
                self.connections[id_, idx, 1] = _[1]  # tail entity
                self.connections[id_, idx, 2] = _[2]  # confidence
        return degrees


class GMUCBaseSampler(GMUCData):
    """Traditional GMUC random sampling mode.
    """
    def __init__(self, args):
        super().__init__(args)

    def generate_false(self, query_triples, candidates):
        """Generate false triples.

        Args:
            query_triples: Query set.
            candidates: All entities that belong to the relationship.

        Returns:
            false_pairs: False triples.(confidence = 0)
            false_left: The head entity of false triples.
            false_right: The tail entity of false triples.
        """
        false_pairs = []
        false_left = []
        false_right = []

        for triple in query_triples:
            for i in range(self.args.num_neg):
                e_h = triple[0]
                rel = triple[1]
                e_t = triple[2]
                while True:
                    noise = random.choice(candidates)
                    if (noise not in self.e1rel_e2[e_h + rel]) and noise != e_t:
                        break
                false_pairs.append([self.symbol2id[e_h], self.symbol2id[noise], 0.0])  # (h, t, s)
                false_left.append(self.ent2id[e_h])
                false_right.append(self.ent2id[noise])

        return false_pairs, false_left, false_right

    def get_train(self):
        return [{k: v} for k, v in self.train_tasks.items()]

    def get_valid(self):
        return [{k: v} for k, v in self.dev_tasks.items()]

    def get_test(self):
        return [{k: v} for k, v in self.test_tasks.items()]



import torch.nn as nn
import torch


class Model(nn.Module):
    def __init__(self, args):
        super(Model, self).__init__()

    def init_emb(self):
        raise NotImplementedError

    def score_func(self, head_emb, relation_emb, tail_emb):
        raise NotImplementedError

    def forward(self, triples, negs, mode):
        raise NotImplementedError

    def tri2emb(self, triples, negs=None, mode="single"):
        """Get embedding of triples.

        This function get the embeddings of head, relation, and tail
        respectively. each embedding has three dimensions.

        Args:
            triples (tensor): This tensor save triples id, which dimension is
                [triples number, 3].
            negs (tensor, optional): This tenosr store the id of the entity to
                be replaced, which has one dimension. when negs is None, it is
                in the test/eval phase. Defaults to None.
            mode (str, optional): This arg indicates that the negative entity
                will replace the head or tail entity. when it is 'single', it
                means that entity will not be replaced. Defaults to 'single'.

        Returns:
            head_emb: Head entity embedding.
            relation_emb: Relation embedding.
            tail_emb: Tail entity embedding.
        """

        # if mode == "single":
        #     if negs is None:
        #         head_emb = self.ent_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]
        #         relation_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
        #         tail_emb = self.ent_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]
        #     else:
        #         head_emb = self.ent_emb(negs[:, 0]).unsqueeze(1)  # [bs, 1, dim]
        #         relation_emb = self.rel_emb(negs[:, 1]).unsqueeze(1)  # [bs, 1, dim]
        #         tail_emb = self.ent_emb(negs[:, 2]).unsqueeze(1)  # [bs, 1, dim]

        if mode == "single":
            head_emb = self.ent_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]
            relation_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            tail_emb = self.ent_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]

        elif mode == "head-batch" or mode == "head_predict":
            if negs is None:
                head_emb = self.ent_emb.weight.data.unsqueeze(0)  # [1, num_ent, dim]
            else:
                head_emb = self.ent_emb(negs)  # [bs, num_neg, dim]

            relation_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            tail_emb = self.ent_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]

        elif mode == "tail-batch" or mode == "tail_predict":
            head_emb = self.ent_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]
            relation_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]

            if negs is None:
                tail_emb = self.ent_emb.weight.data.unsqueeze(0)  # [1, num_ent, dim]
            else:
                tail_emb = self.ent_emb(negs)  # [bs, num_neg, dim]
        return head_emb, relation_emb, tail_emb



import torch
import torch.nn as nn

class BEUrRE_Loss(nn.Module):
    """The loss function of BEUrRE

    Attributes:
        args: Some pre-set parameters, etc
        model: The BEUrRE model for training.
    """
    def __init__(self, args, model):
        super(BEUrRE_Loss, self).__init__()
        self.args = args
        self.model = model

    def get_logic_loss(self, model, ids, args):
        """
        Calculates the logic loss for the model based on transitive and composite rule regularizations.

        Args:
            model: The BEUrRE model instance.
            ids: Tensor of triple ids.
            args: Model configuration parameters including regularization coefficients.

        Returns:
            The logic loss calculated from transitive and composite rule regularizations.
        """
        ids = ids[:, :3].to(torch.long)
        # transitive rule loss regularization
        transitive_coff = torch.tensor(args.regularization['transitive']).to(args.gpu)
        if transitive_coff > 0:
            transitive_rule_reg = transitive_coff * model.transitive_rule_loss(ids)
        else:
            transitive_rule_reg = 0

        # composite rule loss regularization
        composite_coff = torch.tensor(args.regularization['composite']).to(args.gpu)
        if composite_coff > 0:
            composition_rule_reg = composite_coff * model.composition_rule_loss(ids)
        else:
            composition_rule_reg = 0

        return (transitive_rule_reg + composition_rule_reg) / len(ids)


    def main_mse_loss(self, model, ids):
        """
        Computes the Mean Squared Error (MSE) loss for the model.

        Args:
            model: The BEUrRE model instance.
            ids: Tensor of triple ids(the confidence of negative samples is zero).

        Returns:
            mse: The MSE loss for the given triples.
        """
        criterion = nn.MSELoss(reduction='mean')
        prediction = model(ids[:, :3].to(torch.long), train=True).to(torch.float64)

        if ids.shape[1] > 3:
            truth = ids[:, 3]
        else:
            truth = torch.zeros(ids.shape[0], dtype=torch.float64).to(self.args.gpu)

        mse = criterion(prediction, truth)

        return mse


    def L2_regularization(self, model, ids, args):
        """
        Computes the L2 regularization loss for the model.

        Args:
            model: The BEUrRE model instance.
            ids: Tensor of triple ids.
            args: Model configuration parameters including regularization coefficients.

        Returns:
            L2_reg: The L2 regularization loss.
        """
        ids = ids[:, :3].to(torch.long)
        regularization = args.regularization
        device = args.gpu

        # regularization on delta
        delta_coff, min_coff = torch.tensor(regularization['delta']).to(device), torch.tensor(regularization['min']).to(
            device)
        delta_reg1 = delta_coff * torch.norm(torch.exp(model.delta_embedding[ids[:, 0]]), dim=1).mean()
        delta_reg2 = delta_coff * torch.norm(torch.exp(model.delta_embedding[ids[:, 2]]), dim=1).mean()

        min_reg1 = min_coff * torch.norm(model.min_embedding[ids[:, 0]], dim=1).mean()
        min_reg2 = min_coff * torch.norm(model.min_embedding[ids[:, 2]], dim=1).mean()

        rel_trans_coff = torch.tensor(regularization['rel_trans']).to(device)
        rel_trans_reg = rel_trans_coff * (
                torch.norm(torch.exp(model.rel_trans_for_head[ids[:, 1]]), dim=1).mean() + \
                torch.norm(torch.exp(model.rel_trans_for_tail[ids[:, 1]]), dim=1).mean()
        )

        rel_scale_coff = torch.tensor(regularization['rel_scale']).to(device)
        rel_scale_reg = rel_scale_coff * (
                torch.norm(torch.exp(model.rel_scale_for_head[ids[:, 1]]), dim=1).mean() + \
                torch.norm(torch.exp(model.rel_scale_for_tail[ids[:, 1]]), dim=1).mean()
        )

        L2_reg = delta_reg1 + delta_reg2 + min_reg1 + min_reg2 + rel_trans_reg + rel_scale_reg

        return L2_reg


    def forward(self, model, ids, negative_samples, args):
        """
        Calculating the total loss including MSE loss, logic loss, and L2 regularization.

        Args:
            model: The BEUrRE model instance.
            ids: Tensor of positive triple ids.
            negative_samples: Tensor of negative triple ids.
            args: Model configuration parameters including regularization coefficients.

        Returns:
            loss: The total loss combining MSE loss, logic loss, and L2 regularization.
        """
        NEG_RATIO = 1
        pos_loss = self.main_mse_loss(model, ids)

        negative_samples, _ = model.random_negative_sampling(ids[:, :3].to(torch.long), ids[:, 3])

        neg_loss = self.main_mse_loss(model, negative_samples)

        main_loss = pos_loss + NEG_RATIO * neg_loss

        if hasattr(args, 'RULE_CONFIGS'):
            logic_loss = self.get_logic_loss(model, ids, args)
        else:
            logic_loss = 0

        L2_reg = self.L2_regularization(model, ids, args)

        loss = main_loss + L2_reg + logic_loss

        return loss


import torch
import torch.nn.functional as F
import torch.nn as nn


class Margin_Loss(nn.Module):
    """Margin Ranking Loss

    Attributes:
        args: Some pre-set parameters, etc 
        model: The KG model for training.
    """

    def __init__(self, args, model):
        super(Margin_Loss, self).__init__()
        self.args = args
        self.model = model
        self.loss = nn.MarginRankingLoss(self.args.margin)

    def forward(self, pos_score, neg_score):
        """Creates a criterion that measures the loss given inputs pos_score and neg_score. In math:
        
        \text{loss}(x1, x2, y) = \max(0, -y * (x1 - x2) + \text{margin})
        
        Args:
            pos_score: The score of positive samples.
            neg_score: The score of negative samples.
        Returns:
            loss: The training loss for back propagation.
        """
        label = torch.Tensor([1]).type_as(pos_score)
        loss = self.loss(pos_score, neg_score, label)

        return loss



from .BaseLitModel import BaseLitModel
from .UKGELitModel import UKGELitModel
from .BEUrRELitModel import BEUrRELitModel
from .GMUCLitModel import GMUCLitModel
from .UKGEPSLLitModel import UKGEPSLLitModel
from .PASSLEAFLitModel import PASSLEAFLitModel
from .UKGsELitModel import UKGsELitModel
from .GTransELitModel import GTransELitModel
from .UPGATLitModel import UPGATLitModel



import numpy as np
from collections import defaultdict as ddict
from .BaseLitModel import BaseLitModel
from ..eval_task import *
from torch.autograd import Variable


class GMUCLitModel(BaseLitModel):
    """
    Processing of training, evaluation and testing for GMUC and GMUC+.
    """

    def __init__(self, model, train_sampler, args):
        super().__init__(model, args)
        self.sampler = train_sampler
        self.rel2candidates = train_sampler.rel2candidates
        self.symbol2id = train_sampler.symbol2id
        self.ent2id = train_sampler.ent2id
        self.e1rel_e2 = train_sampler.e1rel_e2
        self.rele1_e2 = train_sampler.rele1_e2

    def forward(self, x):
        return self.model(x)

    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=0.1)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser

    def training_step(self, batch, batch_idx):

        support          = batch["support"]
        query            = batch["query"]
        false            = batch["false"]
        support_meta     = batch["support_meta"]
        query_meta       = batch["query_meta"]
        false_meta       = batch["false_meta"]
        symbolid_ic      = batch["symbolid_ic"]

        if self.args.has_ont:
            query_scores, query_scores_var, false_scores, query_confidence = self.model(support, support_meta, query,
                                                                                        query_meta, false, false_meta,
                                                                                        self.args.if_ne)
            loss = self.loss(query_scores, query_scores_var, false_scores, query_confidence, symbolid_ic)
        else:
            (query_scores, query_scores_var, query_ae_loss, false_scores, false_scores_var, false_ae_loss,
             query_confidence) = self.model(support, support_meta, query, query_meta, false, false_meta)
            loss = self.loss(query_scores, query_scores_var, query_ae_loss, false_scores, query_confidence)

        self.log("Train|loss", loss, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        results = self.val_test(batch)

        return results

    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        results = self.val_test(batch)

        return results

    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def val_test(self, batch):
        task = batch["query"]
        test_tasks = batch["triples"]

        support_triples = test_tasks[:self.args.few]  # pos and neg support

        support_pairs = [[self.symbol2id[triple[0]], self.symbol2id[triple[2]], float(triple[3])] for triple in
                         support_triples]
        support_left = [self.ent2id[triple[0]] for triple in support_triples]
        support_right = [self.ent2id[triple[2]] for triple in support_triples]
        support_meta = self.sampler.get_meta(support_left, support_right)

        if self.args.if_GPU:
            support = Variable(torch.Tensor(support_pairs)).cuda()
        else:
            support = Variable(torch.Tensor(support_pairs))

        if self.args.type_constrain:
            candidates = self.rel2candidates[task]
        else:
            candidates = list(self.ent2id.keys())
            candidates = candidates[:1000]

        all_conf, raw_conf_filter, filter_conf_filter, ndcg = [], [], [], []
        pos_mae, pos_mse = [], []
        neg_mae, neg_mse = [], []
        mr, mrr = [], []
        hits = ddict(list)
        r_mr, r_mrr = [], []
        r_hits = ddict(list)

        for e1, e2s in self.rele1_e2[task].items():
            query_pairs = []
            query_left = []
            query_right = []
            query_confidence = []  # len=len(true_e2)

            true_e2 = []
            true_s = []

            for _ in e2s:
                if [e1, task, _[0], _[1]] in support_triples:
                    continue
                true_e2.append(_[0])
                true_s.append(float(_[1]))

            num_e2 = len(true_e2)
            if num_e2 == 0:
                continue

            for i in range(num_e2):
                query_pairs.append([self.symbol2id[e1], self.symbol2id[true_e2[i]]])
                query_left.append(self.ent2id[e1])
                query_right.append(self.ent2id[true_e2[i]])
                query_confidence.append(true_s[i])  # pos queries

            for ent in candidates:
                # if (ent not in self.e1rel_e2[e1 + task]) and (ent not in true_e2):
                query_pairs.append([self.symbol2id[e1], self.symbol2id[ent]])
                query_left.append(self.ent2id[e1])
                query_right.append(self.ent2id[ent])
            query = Variable(torch.LongTensor(query_pairs)).cuda()
            query_meta = self.sampler.get_meta(query_left, query_right)

            # match support set and each query
            if self.args.has_ont:
                scores, scores_var = self.model.score_func(support, support_meta, query, query_meta)
            else:
                scores, scores_var, _ = self.model.score_func(support, support_meta, query, query_meta)

            scores.detach()
            scores = scores.data
            scores = scores.cpu().numpy()
            scores_var.detach()
            scores_var = scores_var.data
            scores_var = scores_var.cpu().numpy()

            f_scores = np.array([])
            f_scores = np.append(f_scores, [scores[i] for i in range(num_e2)])

            for idx, ent in enumerate(candidates):
                if (ent not in self.e1rel_e2[e1 + task]) and (ent not in true_e2):
                    f_scores = np.append(f_scores, scores[idx + num_e2])

            # filter ndcg
            # score_sort = list(np.argsort(f_scores))[::-1]
            # ranks = [score_sort.index(i) + 1 for i in range(num_e2)]  # 所有queries的排名，前num_e2个是pos
            #
            # confidence_arr = np.array(query_confidence)
            # rank_arr = np.array(ranks)
            # discounts = np.log2(rank_arr + 1)
            # discounted_gains = confidence_arr / discounts
            # dcg = np.sum(discounted_gains)
            #
            # confidence_sort = sorted(query_confidence, reverse=True)
            # confidence_sort = np.array(confidence_sort)
            # idcg = np.sum(confidence_sort / np.log2(np.arange(len(confidence_sort)) + 2))
            # ndcg_ = dcg / idcg
            # ndcg.append(ndcg_)

            all_conf.extend(query_confidence)

            # ============================= raw tail entity prediction =============================
            scores_filter = np.array([])
            num_e2_filter = 0

            for i in range(scores.size):
                if i < num_e2:
                    if query_confidence[i] > self.args.confidence_filter:
                        scores_filter = np.append(scores_filter, scores[i])
                        raw_conf_filter.append(query_confidence[i])
                        num_e2_filter += 1
                else:
                    scores_filter = np.append(scores_filter, scores[i])

            score_sort = list(np.argsort(scores_filter))[::-1]
            ranks = [score_sort.index(i) + 1 for i in range(num_e2_filter)]

            ranks_sort = sorted(ranks)
            for i in range(len(ranks)):
                rank = ranks_sort[i] - i
                if rank <= 0:
                    rank = 1
                for k in self.args.calc_hits:
                    r_hits['r_hits@{}'.format(k)].append(1.0 if rank <= k else 0.0)
                r_mrr.append(1.0 / rank)
                r_mr.append(rank)

            # ============================= filter tail entity prediction =============================
            scores_filter = np.array([])
            num_e2_filter = 0

            for i in range(f_scores.size):
                if i < num_e2:
                    if query_confidence[i] > self.args.confidence_filter:
                        scores_filter = np.append(scores_filter, f_scores[i])
                        filter_conf_filter.append(query_confidence[i])
                        num_e2_filter += 1
                else:
                    scores_filter = np.append(scores_filter, f_scores[i])

            score_sort = list(np.argsort(scores_filter))[::-1]
            ranks = [score_sort.index(i) + 1 for i in range(num_e2_filter)]

            ranks_sort = sorted(ranks)
            for i in range(len(ranks)):
                rank = ranks_sort[i] - i
                if rank <= 0:
                    rank = 1
                for k in self.args.calc_hits:
                    hits['hits@{}'.format(k)].append(1.0 if rank <= k else 0.0)
                mrr.append(1.0 / rank)
                mr.append(rank)

            # ==================== confidence prediction ============================
            pos_true_score = np.array(query_confidence)
            pos_pred_score = np.array(scores_var[:num_e2])

            pos_mae.extend(abs(pos_true_score - pos_pred_score))
            pos_mse.extend((pos_true_score - pos_pred_score) ** 2)

            neg_pred_score = np.array(scores_var[num_e2:])
            neg_mae.extend(abs(neg_pred_score))
            neg_mse.extend((neg_pred_score) ** 2)


        all_conf_arr = np.array(raw_conf_filter)
        all_rank_arr = np.array(r_mr)
        r_wmr_ = np.sum(all_conf_arr * all_rank_arr) / np.sum(all_conf_arr)
        all_rr_arr = np.array(r_mrr)
        r_wmrr_ = np.sum(all_conf_arr * all_rr_arr) / np.sum(all_conf_arr)

        all_conf_arr = np.array(filter_conf_filter)
        all_rank_arr = np.array(mr)
        wmr_ = np.sum(all_conf_arr * all_rank_arr) / np.sum(all_conf_arr)
        all_rr_arr = np.array(mrr)
        wmrr_ = np.sum(all_conf_arr * all_rr_arr) / np.sum(all_conf_arr)

        results = dict()
        results["MAE"] = np.mean(pos_mae)
        results["MSE"] = np.mean(pos_mse)
        # raw result
        results["raw_mr"] = np.mean(r_mr) if r_mr else 0
        results["raw_mrr"] = np.mean(r_mrr) if r_mrr else 0
        results["raw_wmr"] = r_wmr_
        results["raw_wmrr"] = r_wmrr_
        for k in self.args.calc_hits:
            results['raw_hits@{}'.format(k)] = np.mean(r_hits['r_hits@{}'.format(k)]) if r_hits['r_hits@{}'.format(k)] else 0
        # filter result
        results["mr"] = np.mean(mr) if mr else 0
        results["mrr"] = np.mean(mrr) if mrr else 0
        # results["ndcg"] = np.mean(ndcg)
        results["wmr"] = wmr_
        results["wmrr"] = wmrr_
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = np.mean(hits['hits@{}'.format(k)]) if hits['hits@{}'.format(k)] else 0

        return results

    def configure_optimizers(self):
        """Setting optimizer and lr_scheduler.

        Returns:
            optim_dict: Record the optimizer and lr_scheduler, type: dict.
        """
        if self.args.has_ont:
            milestones = 10000
            gamma = 0.25
        else:
            milestones = int(self.args.max_epochs / 2)
            gamma = 0.1
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, weight_decay=self.args.weight_decay)
        StepLR = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[milestones], gamma=gamma)
        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict

    def get_results(self, results, mode):
        outputs = ddict(float)
        count = len(results)

        for metric in list(results[0].keys()):
            final_metric = "_".join([mode, metric])
            outputs[final_metric] = np.around(np.array([o[metric] for o in results]).sum() / count, decimals=3).item()

        return outputs



import torch
import os
import torch.nn as nn


def conf_predict(batch, model):
    """The goal of evaluate task is to predict the confidence of triples.

    Args:
        batch: The batch of the triples for validation or test.
        model: The UKG model for training.

    Returns:
        MAE: Mean absolute error.
        MSE: Mean Square error.
    """
    pos_triple = batch["positive_sample"]
    confidence = pos_triple[:, 3]

    pred_score = model.get_score(batch, "single")

    pred_score = pred_score.squeeze()  # 维度压缩
    MAE_loss = nn.L1Loss(reduction="sum")
    # MAE = MAE_loss(pred_score, confidence) * batch["positive_sample"].shape[0]
    MAE = MAE_loss(pred_score, confidence)

    MSE_loss = nn.MSELoss(reduction="sum")
    # MSE = MSE_loss(pred_score, confidence) * batch["positive_sample"].shape[0]
    MSE = MSE_loss(pred_score, confidence)

    return MAE, MSE



# -*- coding: utf-8 -*-
import argparse
import pytorch_lightning as pl
from ..lit_model import BaseLitModel
from ..data import *


def setup_parser():
    """Set up Python's ArgumentParser with data, model, trainer, and other arguments."""
    parser = argparse.ArgumentParser(add_help=False)

    # Add Trainer specific arguments, such as --max_epochs, --gpus, --precision
    trainer_parser = pl.Trainer.add_argparse_args(parser)
    trainer_parser._action_groups[1].title = "Trainer Args"  # pylint: disable=protected-access
    parser = argparse.ArgumentParser(add_help=False, parents=[trainer_parser])

    # Basic arguments
    parser.add_argument('--model_name', default="UKGE", type=str, help='The name of model.')
    parser.add_argument('--dataset_name', default="nl27k", type=str, help='The name of dataset.')
    parser.add_argument('--data_class', default="KGDataModule", type=str,
                        help='The name of data preprocessing module, default KGDataModule.')
    parser.add_argument("--litmodel_name", default="KGELitModel", type=str,
                        help='The name of processing module of training, evaluation and testing, default KGELitModel.')
    parser.add_argument("--train_sampler_class", default="UniSampler", type=str,
                        help='Sampling method used in training, default UniSampler.')
    parser.add_argument("--test_sampler_class", default="TestSampler", type=str,
                        help='Sampling method used in validation and testing, default TestSampler.')
    parser.add_argument('--loss_name', default="Adv_Loss", type=str, help='The name of loss function.')
    parser.add_argument('--negative_adversarial_sampling', '-adv', default=True, action='store_false',
                        help='Use self-adversarial negative sampling.')
    parser.add_argument('--optim_name', default="Adam", type=str, help='The name of optimizer')
    parser.add_argument("--seed", default=321, type=int, help='Random seed.')
    parser.add_argument('--margin', default=12.0, type=float, help='The fixed margin in loss function. ')
    parser.add_argument('--confidence_filter', default=0, type=float,
                        help='Set the confidence threshold in the link prediction, where only samples above  filter_confidence will participate in the link prediction task ')
    parser.add_argument('--adv_temp', default=1.0, type=float,
                        help='The temperature of sampling in self-adversarial negative sampling.')
    parser.add_argument('--emb_dim', default=200, type=int, help='The embedding dimension in KGE model.')
    parser.add_argument('--out_dim', default=200, type=int, help='The output embedding dimmension in some KGE model.')
    parser.add_argument('--num_neg', default=10, type=int,
                        help='The number of negative samples corresponding to each positive sample')
    parser.add_argument('--num_ent', default=None, type=int, help='The number of entity, autogenerate.')
    parser.add_argument('--num_rel', default=None, type=int, help='The number of relation, autogenerate.')
    parser.add_argument('--check_per_epoch', default=5, type=int, help='Evaluation per n epoch of training.')
    parser.add_argument('--early_stop_patience', default=5, type=int,
                        help='If the number of consecutive bad results is n, early stop.')
    parser.add_argument("--num_layers", default=2, type=int, help='The number of layers in some GNN model.')
    parser.add_argument('--regularization', '-r', default=0.0, type=float)
    parser.add_argument("--decoder_model", default=None, type=str, help='The name of decoder model, in some model.')
    parser.add_argument('--eval_task', default="link_prediction", type=str,
                        help='The task of validation, default link_prediction.')
    parser.add_argument("--calc_hits", default=[1, 3, 10], type=lambda s: [int(item) for item in s.split(',')],
                        help='calc hits list')
    parser.add_argument('--filter_flag', default=True, action='store_false', help='Filter in negative sampling.')
    parser.add_argument('--gpu', default='cuda:0', type=str, help='Select the GPU in training, default cuda:0.')
    parser.add_argument("--use_wandb", default=False, action='store_true',
                        help='Use "weight and bias" to record the result.')
    parser.add_argument('--use_weight', default=False, action='store_true', help='Use subsampling weight.')
    parser.add_argument('--checkpoint_dir', default="", type=str, help='The checkpoint model path')
    parser.add_argument('--save_config', default=False, action='store_true', help='Save paramters config file.')
    parser.add_argument('--load_config', default=True, action='store_true', help='Load parametes config file.')
    parser.add_argument('--config_path', default="config/TransE_demo_kg.yaml", type=str, help='The config file path.')
    parser.add_argument('--freq_init', default=4, type=int)
    parser.add_argument('--test_only', default=False, action='store_true')
    parser.add_argument('--shuffle', default=True, action='store_false')
    parser.add_argument('--norm_flag', default=False, action='store_true')

    # parser only for PASSLEAF
    parser.add_argument("--passleaf_score_function", default='DistMult', type=str,
                        help='only on PASSLEAF, choose the score function')
    parser.add_argument("--T_new_semi", default=15, type=int,
                        help='only on PASSLEAF, set the epoch to generate semi-supervised sample')
    parser.add_argument("--T_semi_train", default=20, type=int,
                        help='only on PASSLEAF, set the epoch to use semi-supervised sample to train')
    parser.add_argument("--alpha_PASSLEAF", default=0.02, type=float,
                        help='only on PASSLEAF, set the speed to use semi-supervised sample')
    parser.add_argument("--max_pool", default=10000000, type=int,
                        help='only on PASSLEAF, set the maximum number of triples in the pool')

    # Get data, model, and LitModel specific arguments
    lit_model_group = parser.add_argument_group("LitModel Args")
    BaseLitModel.add_to_argparse(lit_model_group)

    data_group = parser.add_argument_group("Data Args")
    BaseDataModule.add_to_argparse(data_group)

    parser.add_argument("--help", "-h", action="help")

    return parser



import torch
import torch.nn.functional as F
import torch.nn as nn


class UKGE_Loss(nn.Module):

    def __init__(self, args, model):
        super(UKGE_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, pos_sample):
        confidence = pos_sample[:, 3]  # get confidence of each sample
        pos_score = pos_score.squeeze()

        loss_1 = torch.sum((pos_score - confidence) ** 2) # l_pos
        loss_2 = torch.sum(neg_score ** 2) # l_neg

        loss = loss_1 + loss_2 / neg_score.shape[1] # Divide l_neg by num_neg to balance the impact of positive and negative samples on loss
        return loss



from .BaseLitModel import BaseLitModel
from ..eval_task import *


class UPGATLitModel(BaseLitModel):
    """Processing of training, evaluation and testing.
    """

    def __init__(self, model, args):
        super().__init__(model, args)

    def forward(self, x):
        return self.model(x)

    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=0.1)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser

    def training_step(self, batch, batch_idx):
        """Getting samples and training in KG model.

        Args:
            batch: The training data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            loss: The training loss for back propagation.
        """
        adj_matrix = batch["adj_matrix"]
        pos_sample = batch["positive_sample"]
        neg_sample = batch["negative_sample"]
        mode = batch["mode"]
        pos_score = self.model(pos_sample, adj_matrix)
        neg_score = self.model(pos_sample, adj_matrix, neg_sample, mode)
        if self.args.teacher_model:
            loss = self.loss(pos_score, neg_score, pos_sample)
        else:
            pseudo_sample = batch["pseudo_sample"]
            pseudo_score = self.model(pseudo_sample, adj_matrix)
            loss = self.loss(pos_score, neg_score, pos_sample, pseudo_score, pseudo_sample)
        self.log("Train|loss", loss, on_step=False, on_epoch=True)

        return loss

    def validation_step(self, batch, batch_idx):
        """Getting samples and validation in KG model.

        Args:
            batch: The evalutaion data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: MAE, MSE, mrr, mr, wmrr, wmr and hits@1,3,5,10.
        """
        results = dict()

        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"

        confidence = self.args.confidence_filter
        ranks_link_predict = link_predict_filter(batch, self.model, confidence, prediction=prediction)
        ranks_link_predict_raw = link_predict_raw(batch, self.model, confidence, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks_link_predict)
        results["mrr"] = torch.sum(1.0 / ranks_link_predict).item()
        results["mr"] = torch.sum(ranks_link_predict).item()
        results["raw_count_for_link"] = torch.numel(ranks_link_predict_raw)
        results["raw_mrr"] = torch.sum(1.0 / ranks_link_predict_raw).item()
        results["raw_mr"] = torch.sum(ranks_link_predict_raw).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks_link_predict[ranks_link_predict <= k])
            results['raw_hits@{}'.format(k)] = torch.numel(ranks_link_predict_raw[ranks_link_predict_raw <= k])

        pos_triple = batch["positive_sample"]
        mask = pos_triple[:, -1] >= confidence
        if prediction == "all":
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        else:
            conf = batch['positive_sample'][:, 3]
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        return results

    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        """Getting samples and test in KG model.

        Args:
            batch: The test data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: MAE, MSE, mrr, mr, wmrr, wmr and hits@1,3,5,10.
        """
        results = dict()

        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"
        confidence = self.args.confidence_filter
        ranks_link_predict = link_predict_filter(batch, self.model, confidence, prediction=prediction)
        ranks_link_predict_raw = link_predict_raw(batch, self.model, confidence, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks_link_predict)
        results["mrr"] = torch.sum(1.0 / ranks_link_predict).item()
        results["mr"] = torch.sum(ranks_link_predict).item()
        results["raw_count_for_link"] = torch.numel(ranks_link_predict_raw)
        results["raw_mrr"] = torch.sum(1.0 / ranks_link_predict_raw).item()
        results["raw_mr"] = torch.sum(ranks_link_predict_raw).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks_link_predict[ranks_link_predict <= k])
            results['raw_hits@{}'.format(k)] = torch.numel(ranks_link_predict_raw[ranks_link_predict_raw <= k])

        pos_triple = batch["positive_sample"]
        mask = pos_triple[:, -1] >= confidence
        if prediction == "all":
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        else:
            conf = batch['positive_sample'][:, 3]
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        return results

    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def configure_optimizers(self):
        """Setting optimizer and lr_scheduler.

        Returns:
            optim_dict: Record the optimizer and lr_scheduler, type: dict.
        """
        milestones = int(self.args.max_epochs / 8)
        beta1 = 0.9  # set β1
        beta2 = 0.99  # set β2
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, betas=(beta1, beta2))
        StepLR = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[milestones], gamma=0.1)
        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict


import torch.nn as nn
import torch
from .model import Model


class PASSLEAF(Model):
    """`A Pool-bAsed Semi-Supervised LEArning Framework for Uncertain Knowledge Graph Embedding`_ (PASSLEAF).

    Attributes:
        args: Model configuration parameters.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        w: Weight when calculate confidence scores
        b: Bias when calculate confidence scores
    .. _Translating Embeddings for Modeling Multi-relational Data: http://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-rela
    """

    def __init__(self, args):
        super(PASSLEAF, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.w = torch.nn.Parameter(torch.tensor(1.0), requires_grad=True)
        self.b = torch.nn.Parameter(torch.tensor(0.0), requires_grad=True)

        self.init_emb()

    def init_emb(self):
        """
            Initialize the entity and relation embeddings in the form of a uniform distribution.
        """
        model = self.args.passleaf_score_function
        if model == 'DistMult':
            "`Embedding Entities and Relations for Learning and Inference in Knowledge Bases`_ (DistMult)"
            self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
            self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
            nn.init.xavier_uniform_(self.ent_emb.weight.data)
            nn.init.xavier_uniform_(self.rel_emb.weight.data)
        elif model == 'ComplEx':
            "`Complex Embeddings for Simple Link Prediction`_ (ComplEx)"
            self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim * 2)
            self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim * 2)
            nn.init.xavier_uniform_(self.ent_emb.weight.data)
            nn.init.xavier_uniform_(self.rel_emb.weight.data)
        else: #RotatE
            "`RotatE: Knowledge Graph Embedding by Relational Rotation in Complex Space`_ (RotatE)"
            """Initialize the entity and relation embeddings in the form of a uniform distribution."""
            # print(self.gama)
            """Initialize the entity and relation embeddings in the form of a uniform distribution."""
            self.epsilon = 2.0
            self.margin = nn.Parameter(
                torch.Tensor([self.args.margin]),
                requires_grad=False
            )
            self.embedding_range = nn.Parameter(
                torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]),
                requires_grad=False
            )
            self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim * 2)
            self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
            nn.init.xavier_uniform_(self.ent_emb.weight.data)
            nn.init.xavier_uniform_(self.rel_emb.weight.data)


    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        The formula for calculating the score of DistMult is :math:`\transpose{h} \operatorname{diag}(r) t`.

        The formula for calculating the score of ComplEx is :math:` \Re(\transpose{h} \operatorname{diag}(r) \overline{t})`.

        The formula for calculating the score of RotatE is :math:`\gamma - \|h \circ r - t\|`

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """

        model = self.args.passleaf_score_function # use different score function according to args.passleaf_score_function

        if model == 'DistMult':
            if mode == 'head-batch':
                score = head_emb * (relation_emb * tail_emb)
            else:
                score = (head_emb * relation_emb) * tail_emb

            score = score.sum(dim=-1)
            score = score.to(self.args.gpu)

            """ 1.Bounded rectifier """
            # shape = score.shape
            # tmp_max = torch.max(self.w * score + self.b, torch.zeros(shape, device=self.args.gpu))
            # score = torch.min(tmp_max, torch.ones(shape, device=self.args.gpu))

            """ 2.Logistic function"""
            score = torch.sigmoid(self.w * score + self.b) # use UKGE_logi in PASSLEAF


        if model == 'ComplEx':
            re_head, im_head = torch.chunk(head_emb, 2, dim=-1)
            re_relation, im_relation = torch.chunk(relation_emb, 2, dim=-1)
            re_tail, im_tail = torch.chunk(tail_emb, 2, dim=-1)

            score = torch.sum(
                re_head * re_tail * re_relation
                + im_head * im_tail * re_relation
                + re_head * im_tail * im_relation
                - im_head * re_tail * im_relation,
                -1
            )
            score = torch.sigmoid(self.w * score + self.b)

        if model == 'RotatE':
            pi = 3.14159265358979323846
            re_head, im_head = torch.chunk(head_emb, 2, dim=-1)
            re_tail, im_tail = torch.chunk(tail_emb, 2, dim=-1)

            # Make phases of relations uniformly distributed in [-pi, pi]

            phase_relation = relation_emb / (self.embedding_range.item() / pi)

            re_relation = torch.cos(phase_relation)
            im_relation = torch.sin(phase_relation)

            if mode == 'head-batch':
                re_score = re_relation * re_tail + im_relation * im_tail
                im_score = re_relation * im_tail - im_relation * re_tail
                re_score = re_score - re_head
                im_score = im_score - im_head
            else:
                re_score = re_head * re_relation - im_head * im_relation
                im_score = re_head * im_relation + im_head * re_relation
                re_score = re_score - re_tail
                im_score = im_score - im_tail

            score = torch.stack([re_score, im_score], dim=0)
            score = score.norm(dim=0)
            score = self.margin.item() - score.sum(dim=-1)
            score = torch.sigmoid(self.w * score + self.b)

        return score

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        triples = triples[:, :3].to(torch.int)
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        triples = triples[:, :3].to(torch.int)

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score



import torch.nn as nn
import torch
from .model import Model


class FocusE(Model):
    """`Learning Embeddings from Knowledge Graphs With Numeric Edge Attributes`_ (FocusE), which injects
    numeric edge attributes into the scoring layer of a traditional knowledge graph embedding architecture.

    Attributes:
        args: Model configuration parameters.
        epsilon: Calculate embedding_range.
        margin: Calculate embedding_range and loss.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        base_model: The based traditional knowledge graph embedding method.
        beta: the structural influence

    .. _Learning Embeddings from Knowledge Graphs With Numeric Edge Attributes: https://www.ijcai.org/proceedings/2021/395
    """

    def __init__(self, args):
        super(FocusE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.base_model = args.base_model
        self.beta = 1.0

        self.init_emb()

    def init_emb(self):
        """Initialize the entity and relation embeddings in the form of a uniform distribution.

        """
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]),
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]),
            requires_grad=False
        )

        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples using different traditional knowledge graph embedding method.

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """

        if self.base_model == "DistMult":
            if mode == 'head-batch':
                score = head_emb * (relation_emb * tail_emb)
            else:
                score = (head_emb * relation_emb) * tail_emb

            score = score.sum(dim=-1)
        elif self.base_model == "ComplEX":
            re_head, im_head = torch.chunk(head_emb, 2, dim=-1)
            re_relation, im_relation = torch.chunk(relation_emb, 2, dim=-1)
            re_tail, im_tail = torch.chunk(tail_emb, 2, dim=-1)

            score = torch.sum(re_head * re_tail * re_relation +
                              im_head * im_tail * re_relation +
                              re_head * im_tail * im_relation -
                              im_head * re_tail * im_relation, -1
                              )
        else:  # 暂时默认TransE
            score = (head_emb + relation_emb) - tail_emb
            score = self.margin.item() - torch.norm(score, p=1, dim=-1)

        return score

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The quadruple ids, as (h, r, t, c), shape:[batch_size, r].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        w = triples[:, 3]
        triples = triples[:, :3].to(torch.int)
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        softplus = nn.Softplus()
        score = softplus(score)

        if negs is None:
            alpha = self.beta + (torch.ones(w.shape).to(self.args.gpu) - w) * (1 - self.beta)
        else:
            alpha = self.beta + w * (1 - self.beta)

        alpha = alpha.unsqueeze(1)
        score = alpha * score

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples(But not the confidence).
        """
        triples = batch["positive_sample"]
        triples = triples[:, :3].to(torch.int)
        triples = triples.to(self.args.gpu)

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        softplus = nn.Softplus()
        score = softplus(score)

        return score


    def adjust_parameters(self, current_epoch):
        """To adjust the structural influence beta

        Args:
            current_epoch: Which epoch is currently being trained
        """
        self.beta = max(0.0, 1.0 - (current_epoch / (self.args.max_epochs-1)))
        print(self.beta)


#!/usr/bin/env python
# coding: utf-8

import setuptools
import os


with open("README.md", 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='unKR',
    version='0.8.4',
    author='SEUKGE',
    author_email='slchen@seu.edu.cn',
    url='https://github.com/CodeSlogan/unKR',
    description='An Open Source Library for uncertain Knowledge Reasoning',
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'pytorch_lightning==1.5.10',
        'PyYAML',
        'wandb',
        'IPython',
        'gensim'
    ],
    python_requires=">=3.6"
)



import torch
import torch.nn.functional as F
import torch.nn as nn


class UKGsE_Loss(nn.Module):

    def __init__(self, args, model):
        super(UKGsE_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, pos_sample):
        confidence = pos_sample[:, 3]  # 拿到每个三元组的置信度
        pos_score = pos_score.squeeze()

        loss_1 = torch.mean((pos_score - confidence) ** 2)
        loss_2 = torch.mean((neg_score - 1e-08) ** 2)   # MSE

        loss = loss_1 + loss_2
        return loss



import torch
import torch.nn.functional as F
import torch.nn as nn

class Adv_Loss(nn.Module):
    """Negative sampling loss with self-adversarial training. 

    Attributes:
        args: Some pre-set parameters, such as self-adversarial temperature, etc. 
        model: The KG model for training.
    """
    def __init__(self, args, model):
        super(Adv_Loss, self).__init__()
        self.args = args
        self.model = model
    def forward(self, pos_score, neg_score, subsampling_weight=None):
        """Negative sampling loss with self-adversarial training. In math:
        
        L=-\log \sigma\left(\gamma-d_{r}(\mathbf{h}, \mathbf{t})\right)-\sum_{i=1}^{n} p\left(h_{i}^{\prime}, r, t_{i}^{\prime}\right) \log \sigma\left(d_{r}\left(\mathbf{h}_{i}^{\prime}, \mathbf{t}_{i}^{\prime}\right)-\gamma\right)
        
        Args:
            pos_score: The score of positive samples.
            neg_score: The score of negative samples.
            subsampling_weight: The weight for correcting pos_score and neg_score.

        Returns:
            loss: The training loss for back propagation.
        """

        if self.args.negative_adversarial_sampling:
            neg_score = (F.softmax(neg_score * self.args.adv_temp, dim=1).detach()
                        * F.logsigmoid(-neg_score)).sum(dim=1)  #shape:[bs]
        else:
            neg_score = F.logsigmoid(-neg_score).mean(dim = 1)

        pos_score = F.logsigmoid(pos_score).view(neg_score.shape[0]) #shape:[bs]
        # from IPython import embed;embed();exit()

        if self.args.use_weight:
            positive_sample_loss = - (subsampling_weight * pos_score).sum()/subsampling_weight.sum()
            negative_sample_loss = - (subsampling_weight * neg_score).sum()/subsampling_weight.sum()
        else:
            positive_sample_loss = - pos_score.mean()
            negative_sample_loss = - neg_score.mean()

        loss = (positive_sample_loss + negative_sample_loss) / 2

        # if self.args.model_name == 'ComplEx' or self.args.model_name == 'DistMult' or self.args.model_name == 'BoxE' or self.args.model_name=="IterE":
        #     #Use L3 regularization for ComplEx and DistMult
        #     regularization = self.args.regularization * (
        #         self.model.ent_emb.weight.norm(p = 3)**3 + \
        #         self.model.rel_emb.weight.norm(p = 3)**3
        #     )
        #     # embed();exit()
        #     loss = loss + regularization

        regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p=3) ** 3 + self.model.rel_emb.weight.norm(p=3) ** 3
        )
        loss = loss + regularization

        return loss
    
    def normalize(self):
        """calculating the regularization.
        """
        regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p = 3)**3 + \
                self.model.rel_emb.weight.norm(p = 3)**3
            )
        return regularization


import time
import torch
import os
import numpy as np
from tqdm import tqdm


def mean_ndcg(hr_map, model, ent_num):

    ndcg_sum = 0  # nDCG with linear gain
    exp_ndcg_sum = 0  # nDCG with exponential gain
    count = 0

    for h in tqdm(hr_map, desc="Testing_nDCG: "):
        for r in hr_map[h]:
            tw_dict = hr_map[h][r]  # {h:{r:{t:w}}}
            tw_truth = [IndexScore(t, w) for t, w in tw_dict.items()]
            tw_truth.sort(reverse=True)  # descending on w
            ndcg, exp_ndcg = nDCG(h, r, tw_truth, model, ent_num)  # nDCG with linear gain and exponential gain
            # print(ndcg, exp_ndcg)
            ndcg_sum += ndcg
            exp_ndcg_sum += exp_ndcg
            count += 1

    return ndcg_sum / count, exp_ndcg_sum / count


def nDCG(h, r, tw_truth, model, ent_num):
    """
    Compute nDCG(normalized discounted cummulative gain)
    sum(score_ground_truth / log2(rank+1)) / max_possible_dcg
    :param tw_truth: [IndexScore1, IndexScore2, ...], soreted by IndexScore.score descending
    :return:
    """
    # prediction
    ts = [tw.index for tw in tw_truth]             # 真实的排名       [25,60,75,3,7]
    ranks = get_t_ranks(h, r, ts, model, ent_num)  # 根据模型给出的排序 [24,68,222,1,6]

    # linear gain
    gains = np.array([tw.score for tw in tw_truth])  # 从大到小排序好的
    discounts = np.log2(ranks + 1)
    discounted_gains = gains / discounts
    dcg = np.sum(discounted_gains)  # discounted cumulative gain
    # normalize 理想中的排序
    idcg = np.sum(gains / np.log2(np.arange(len(gains)) + 2))  # when ranks = [1, 2, ...len(truth)]
    ndcg = dcg / idcg  # normalized discounted cumulative gain

    # exponential gain
    exp_gains = np.array([2 ** tw.score - 1 for tw in tw_truth])
    exp_discounted_gains = exp_gains / discounts
    exp_dcg = np.sum(exp_discounted_gains)
    # normalize
    exp_idcg = np.sum(
        exp_gains / np.log2(np.arange(len(exp_gains)) + 2))  # when ranks = [1, 2, ...len(truth)]
    exp_ndcg = exp_dcg / exp_idcg  # normalized discounted cumulative gain

    return ndcg, exp_ndcg


def get_t_ranks(h, r, ts, model, ent_num):
    """
    Given some t index, return the ranks for each t
    :return:
    """
    # prediction
    # scores = get_score(h, r, ts, model)  # 对当前的hr{t:w}匹配对进行预测

    N = ent_num   # 预测所有的hrt_1...N
    score_list = get_score(h, r, range(N), model)

    # way1:
    # ranks = np.ones(len(ts), dtype=int)  # initialize rank as all 1
    # for i in range(N):  # compute scores for all concept vectors as t
    #     score_i = score_list[i]
    #     rankplus = (scores < score_i).astype(int)  # rank+1 if score<score_i
    #     ranks += rankplus
    # print("ranks1:", ranks)

    # 使用argsort函数得到排序后的索引
    rank_indices = list(np.argsort(score_list)[::-1])
    ranks = [rank_indices.index(i) + 1 for i in ts]
    ranks = np.array(ranks)
    # print("ranks2:", ranks)

    return ranks


def get_score(h, r, ts, model):
    batch = {}
    # triples = [(h, r, t) for t in ts]
    triples = [(h, r, t) for t in ts]
    batch["positive_sample"] = torch.tensor(np.array(triples))
    scores = model.get_score(batch, "single")
    scores = scores.squeeze()
    scores = scores.cpu()
    scores = scores.detach().numpy()

    return scores


class IndexScore:
    """
    The score of a tail when h and r is given.
    It's used in the ranking task to facilitate comparison and sorting.
    Print w as 3 digit precision float.
    """

    def __init__(self, index, score):
        self.index = index
        self.score = score

    def __lt__(self, other):
        return self.score < other.score

    def __repr__(self):
        # return "(index: %d, w:%.3f)" % (self.index, self.score)
        return "(%d, %.3f)" % (self.index, self.score)

    def __str__(self):
        return "(index: %d, w:%.3f)" % (self.index, self.score)


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import uniform
import numpy as np


class BEUrRE(nn.Module):
    """`Probabilistic Box Embeddings for Uncertain Knowledge Graph Reasoning`_ (BEUrRE), which utilizes the geometry of boxes
    to calibrate probabilistic semantics and facilitating the incorporation of relational constraints.

    Attributes:
        args: Model configuration parameters.
        min_init_value: Initial value range for the minimum boundary of entity boxes. This range is used for uniformly initializing the minimum boundary of each entity box.
        delta_init_value: Initial value range for the size change (delta) of entity boxes. This range is used for uniformly initializing the size change of each entity box.
        min_embedding: Embedding for the minimum boundary of entities, initialized using min_init_value
        delta_embedding: Embedding for the size change (delta) of entity boxes, initialized using delta_init_value
        rel_trans_for_head: Relation-specific translation parameters for head entities
        rel_scale_for_head: Relation-specific scaling parameters for head entities
        rel_trans_for_tail: Relation-specific translation parameters for tail entities
        rel_scale_for_tail: Relation-specific scaling parameters for tail entities
        true_head, true_tail: Variables used for filtering negative samples to ensure that generated negative samples are not true triples.
        gumbel_beta: The beta parameter for the Gumbel distribution, controlling the softness of intersections in box embeddings.
        device: Specifies the device for model training, such as GPU.
        ratio: The ratio of negative samples to positive samples, used for generating negative samples during training.
        vocab_size: Total number of entities
        alpha: A small numeric value to prevent division by zero in calculations.
        clamp_min, clamp_max: Used to limit the range of gradients and embedding values to prevent numeric overflow or excessively large values.
        REL_VOCAB_SIZE: Total number of relations

    .. _Probabilistic Box Embeddings for Uncertain Knowledge Graph Reasoning: https://aclanthology.org/2021.naacl-main.68
    """
    def __init__(self, args, min_init_value=None, delta_init_value=None):
        super(BEUrRE, self).__init__()

        if delta_init_value is None:
            delta_init_value = [-0.1, -0.001]
        if min_init_value is None:
            min_init_value = [1e-4, 0.01]
        self.euler_gamma = 0.57721566490153286060
        self.min_init_value = min_init_value
        self.delta_init_value = delta_init_value

        min_embedding = self.init_embedding(args.num_ent, args.emb_dim, min_init_value)
        delta_embedding = self.init_embedding(args.num_ent, args.emb_dim, delta_init_value)
        self.min_embedding = nn.Parameter(min_embedding)
        self.delta_embedding = nn.Parameter(delta_embedding)

        rel_trans_for_head = torch.empty(args.num_rel, args.emb_dim)
        rel_scale_for_head = torch.empty(args.num_rel, args.emb_dim)
        torch.nn.init.normal_(rel_trans_for_head, mean=0, std=1e-4)  # 1e-4 before
        torch.nn.init.normal_(rel_scale_for_head, mean=1, std=0.2)  # 0.2 before

        rel_trans_for_tail = torch.empty(args.num_rel, args.emb_dim)
        rel_scale_for_tail = torch.empty(args.num_rel, args.emb_dim)
        torch.nn.init.normal_(rel_trans_for_tail, mean=0, std=1e-4)
        torch.nn.init.normal_(rel_scale_for_tail, mean=1, std=0.2)

        # make nn.Parameter
        self.rel_trans_for_head, self.rel_scale_for_head = nn.Parameter(rel_trans_for_head.to(args.gpu)), nn.Parameter(
            rel_scale_for_head.to(args.gpu))
        self.rel_trans_for_tail, self.rel_scale_for_tail = nn.Parameter(rel_trans_for_tail.to(args.gpu)), nn.Parameter(
            rel_scale_for_tail.to(args.gpu))

        self.true_head, self.true_tail = None, None  # for negative sample filtering
        self.gumbel_beta = args.GUMBEL_BETA
        self.device = args.gpu
        self.ratio = args.num_neg
        self.vocab_size = args.num_ent
        self.alpha = 1e-16
        self.clamp_min = 0.0
        self.clamp_max = 1e10
        self.REL_VOCAB_SIZE = args.num_rel
        self.args = args


    def forward(self, ids, train=True):
        """The functions used in the training phase

        Args:
            ids: The quadruples ids, as (h, r, t, c), shape:[batch_size, 4].
            train: mode

        Returns:
            pos_predictions: The confidence score of triples.
        """
        head_boxes = self.transform_head_boxes(ids)
        tail_boxes = self.transform_tail_boxes(ids)

        intersection_boxes = self.intersection(head_boxes, tail_boxes)

        log_intersection = self.log_volumes(intersection_boxes)

        # condition on subject or object
        log_prob = log_intersection - self.log_volumes(tail_boxes)

        pos_predictions = torch.exp(log_prob)
        return pos_predictions

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            scores: The confidence score of triples.
        """
        triples = batch["positive_sample"].to(torch.long)
        triples = triples[:, :3]

        if mode == "tail_predict":
            scores = []

            # Repeating head-entities and relations to construct all possible triples
            head_entities = triples[:, 0].unsqueeze(1).repeat(1, self.vocab_size)
            relations = triples[:, 1].unsqueeze(1).repeat(1, self.vocab_size)
            tail_entities = torch.arange(self.vocab_size, dtype=torch.long).unsqueeze(0).repeat(len(triples), 1).to(self.args.gpu)

            bs_triples = torch.stack([head_entities, relations, tail_entities], dim=2)
            for i in bs_triples:
                score = self.forward(i)
                scores.append(score)

            scores = torch.stack(scores, dim=0)

            return scores

        elif mode == "head_predict":
            scores = []

            # Repeating tail-entities and relations to construct all possible triples
            head_entities = torch.arange(self.vocab_size, dtype=torch.long).unsqueeze(0).repeat(len(triples), 1).to(self.args.gpu)
            relations = triples[:, 1].unsqueeze(1).repeat(1, self.vocab_size)
            tail_entities = triples[:, 2].unsqueeze(1).repeat(1, self.vocab_size)

            bs_triples = torch.stack([head_entities, relations, tail_entities], dim=2)
            for i in bs_triples:
                score = self.forward(i)
                scores.append(score)

            scores = torch.stack(scores, dim=0)

            return scores

        else:
            return self.forward(triples)

    def transform_head_boxes(self, ids):
        """
        Transforms head entity boxes based on the relation-specific parameters.

        Args:
            ids: The quadruples ids, as (h, r, t, c), shape:[batch_size, 4].

        Returns:
            head_boxes: Transformed head entity boxes after applying relation-specific affine transformations.
        """
        head_boxes = self.get_entity_boxes(ids[:, 0])

        rel_ids = ids[:, 1]
        relu = nn.ReLU()

        translations = self.rel_trans_for_head[rel_ids]
        scales = relu(self.rel_scale_for_head[rel_ids])

        # affine transformation
        head_boxes.min_embed += translations
        head_boxes.delta_embed *= scales
        head_boxes.max_embed = head_boxes.min_embed + head_boxes.delta_embed

        return head_boxes

    def transform_tail_boxes(self, ids):
        """
        Transforms tail entity boxes based on the relation-specific parameters.

        Args:
            ids: The quadruples ids, as (h, r, t, c), shape:[batch_size, 4].

        Returns:
            tail_boxes: Transformed tail entity boxes after applying relation-specific affine transformations.
        """
        tail_boxes = self.get_entity_boxes(ids[:, 2])

        rel_ids = ids[:, 1]
        relu = nn.ReLU()

        translations = self.rel_trans_for_tail[rel_ids]
        scales = relu(self.rel_scale_for_tail[rel_ids])

        # affine transformation
        tail_boxes.min_embed += translations
        tail_boxes.delta_embed *= scales
        tail_boxes.max_embed = tail_boxes.min_embed + tail_boxes.delta_embed

        return tail_boxes


    def intersection(self, boxes1, boxes2):
        """
        Computes the intersection of two sets of boxes using the Gumbel softmax trick.

        Args:
            boxes1: The first set of entity boxes.
            boxes2: The second set of entity boxes.

        Returns:
            intersection_box: The intersection of the two sets of boxes.
        """
        intersections_min = self.gumbel_beta * torch.logsumexp(
            torch.stack((boxes1.min_embed / self.gumbel_beta, boxes2.min_embed / self.gumbel_beta)),
            0
        )
        intersections_min = torch.max(
            intersections_min,
            torch.max(boxes1.min_embed, boxes2.min_embed)
        )
        intersections_max = - self.gumbel_beta * torch.logsumexp(
            torch.stack((-boxes1.max_embed / self.gumbel_beta, -boxes2.max_embed / self.gumbel_beta)),
            0
        )
        intersections_max = torch.min(
            intersections_max,
            torch.min(boxes1.max_embed, boxes2.max_embed)
        )

        intersection_box = Box(intersections_min, intersections_max)
        return intersection_box

    def log_volumes(self, boxes, temp=1., gumbel_beta=1., scale=1.):
        """
        Calculates the logarithm of the volumes of boxes.

        Args:
            boxes: A set of entity boxes.
            temp: The temperature parameter for the softplus function.
            gumbel_beta: The beta parameter for the Gumbel distribution.
            scale: Scaling factor for the volumes.

        Returns:
            log_vol: The logarithm of the volumes of the given boxes.
        """
        eps = torch.finfo(boxes.min_embed.dtype).tiny  # type: ignore

        if isinstance(scale, float):
            s = torch.tensor(scale)
        else:
            s = scale

        log_vol = torch.sum(
            torch.log(
                F.softplus(boxes.delta_embed - 2 * self.euler_gamma * self.gumbel_beta, beta=temp).clamp_min(eps)
            ),
            dim=-1
        ) + torch.log(s)

        return log_vol

    def get_entity_boxes(self, entities):
        """
        Retrieves entity boxes based on entity indices.

        Args:
            entities: A tensor containing indices of entities.

        Returns:
            boxes: A Box object representing the boxes of the specified entities.
        """
        min_rep = self.min_embedding[entities]  # batchsize * embedding_size
        delta_rep = self.delta_embedding[entities]
        max_rep = min_rep + torch.exp(delta_rep)
        boxes = Box(min_rep, max_rep)
        return boxes

    def init_embedding(self, vocab_size, embed_dim, init_value):
        """
        Initializes embeddings uniformly within a specified range.

        Args:
            vocab_size: The number of unique entities.
            embed_dim: The dimension of the embeddings.
            init_value: A tuple indicating the range for uniform initialization.

        Returns:
            box_embed: Initialized embeddings.
        """
        distribution = uniform.Uniform(init_value[0], init_value[1])
        box_embed = distribution.sample((vocab_size, embed_dim))
        return box_embed

    def random_negative_sampling(self, positive_samples, pos_probs, neg_per_pos=None):
        """
        Generates negative samples by corrupting either the head or the tail of positive samples.

        Args:
            positive_samples: Tensor of positive samples.
            pos_probs: confidence score of positive samples.
            neg_per_pos: Number of negative samples to generate per positive sample.

        Returns:
            negative_samples: Tensor of generated negative samples.
            neg_probs: confidence score negative samples (0).
        """
        if neg_per_pos is None:
            neg_per_pos = self.args.num_neg
        negative_samples1 = torch.repeat_interleave(positive_samples, neg_per_pos, dim=0)
        negative_samples2 = torch.repeat_interleave(positive_samples, neg_per_pos, dim=0)

        corrupted_heads = [self.get_negative_samples_for_one_positive(pos, neg_per_pos, mode='corrupt_head') for pos in positive_samples]
        corrupted_tails = [self.get_negative_samples_for_one_positive(pos, neg_per_pos, mode='corrupt_tail') for pos in positive_samples]

        negative_samples1[:, 0] = torch.cat(corrupted_heads)
        negative_samples2[:, 2] = torch.cat(corrupted_tails)
        negative_samples = torch.cat((negative_samples1, negative_samples2), 0).to(self.args.gpu)
        neg_probs = torch.zeros(negative_samples.shape[0], dtype=pos_probs.dtype).to(self.args.gpu)

        return negative_samples, neg_probs

    def random_negative_sampling0(self, positive_samples, pos_probs, neg_per_pos=None):
        """
        Generates negative samples randomly without checking against true triples.

        Args:
            positive_samples: Tensor of positive samples.
            pos_probs: confidence score of positive samples.
            neg_per_pos: Number of negative samples to generate per positive sample.

        Returns:
            negative_samples: Tensor of generated negative samples.
            neg_probs: confidence score negative samples (0).
        """
        if neg_per_pos is None:
            neg_per_pos = self.args.num_neg
        negative_samples1 = torch.repeat_interleave(positive_samples, neg_per_pos, dim=0)
        negative_samples2 = torch.repeat_interleave(positive_samples, neg_per_pos, dim=0)

        # corrupt tails
        corrupted_heads = torch.randint(self.vocab_size, (negative_samples1.shape[0],)).to(self.args.gpu)
        corrupted_tails = torch.randint(self.vocab_size, (negative_samples1.shape[0],)).to(self.args.gpu)

        #filter
        bad_heads_idxs = (corrupted_heads == negative_samples1[:,0])
        bad_tails_idxs = (corrupted_tails == negative_samples2[:,2])
        corrupted_heads[bad_heads_idxs] = torch.randint(self.vocab_size, (torch.sum(bad_heads_idxs),)).to(self.device)
        corrupted_tails[bad_tails_idxs] = torch.randint(self.vocab_size, (torch.sum(bad_tails_idxs),)).to(self.device)

        negative_samples1[:, 0] = corrupted_heads
        negative_samples2[:, 2] = corrupted_tails
        negative_samples = torch.cat((negative_samples1, negative_samples2), 0).to(self.args.gpu)
        neg_probs = torch.zeros(negative_samples.shape[0], dtype=pos_probs.dtype).to(self.args.gpu)

        return negative_samples, neg_probs

    def get_negative_samples_for_one_positive(self, positive_sample, neg_per_pos, mode):
        """
        Generates negative samples for one positive sample by corrupting head or tail.

        Args:
            positive_sample: A single positive sample.
            neg_per_pos: Number of negative samples to generate.
            mode: 'corrupt_head' or 'corrupt_tail' indicating which part of the triple to corrupt.

        Returns:
            negative_sample: Generated negative samples for the given positive sample.
        """
        head, relation, tail = positive_sample
        negative_sample_list = []
        negative_sample_size = 0
        while negative_sample_size < neg_per_pos:
            negative_sample = np.random.randint(self.args.num_ent, size=neg_per_pos * 2)

            # filter true values
            if mode == 'corrupt_head' and (int(relation), int(tail)) in self.true_head:  # filter true heads
                # For test data, some (relation, tail) pairs may be unseen and not in self.true_head
                mask = np.in1d(
                    negative_sample,
                    self.true_head[(int(relation), int(tail))],
                    assume_unique=True,
                    invert=True
                )
                negative_sample = negative_sample[mask]
            elif mode == 'corrupt_tail' and (int(head), int(relation)) in self.true_tail:
                mask = np.in1d(
                    negative_sample,
                    self.true_tail[(int(head), int(relation))],
                    assume_unique=True,
                    invert=True
                )
                negative_sample = negative_sample[mask]
            negative_sample_list.append(negative_sample)
            negative_sample_size += negative_sample.size

        negative_sample = np.concatenate(negative_sample_list)[:neg_per_pos]

        negative_sample = torch.from_numpy(negative_sample)
        return negative_sample


    def head_transformation(self, head_boxes, rel_ids):
        """
        Applies affine transformation to head boxes based on the relation.

        Args:
            head_boxes: Box objects for head entities.
            rel_ids: Relation indices corresponding to the head entities.

        Returns:
            head_boxes: Transformed Box objects for head entities.
        """
        relu = nn.ReLU()
        translations = self.rel_trans_for_head[rel_ids]
        scales = relu(self.rel_scale_for_head[rel_ids])
        # affine transformation
        head_boxes.min_embed += translations
        head_boxes.delta_embed *= scales
        head_boxes.max_embed = head_boxes.min_embed + head_boxes.delta_embed

        return head_boxes

    def tail_transformation(self, tail_boxes, rel_ids):
        """
        Applies affine transformation to tail boxes based on the relation.

        Args:
            tail_boxes: Box objects for tail entities.
            rel_ids: Relation indices corresponding to the tail entities.

        Returns:
            tail_boxes: Transformed Box objects for tail entities.
        """
        relu = nn.ReLU()
        translations = self.rel_trans_for_tail[rel_ids]
        scales = relu(self.rel_scale_for_tail[rel_ids])
        # affine transformation
        tail_boxes.min_embed += translations
        tail_boxes.delta_embed *= scales
        tail_boxes.max_embed = tail_boxes.min_embed + tail_boxes.delta_embed

        return tail_boxes

    def get_entity_boxes_detached(self, entities):
        """
        For logic constraint. We only want to optimize relation parameters, so detach entity parameters

        Args:
            entities: A tensor containing indices of entities.

        Returns:
            boxes: A Box object representing the boxes of the specified entities without attached gradients.
        """
        min_rep = self.min_embedding[entities].detach()
        delta_rep = self.delta_embedding[entities].detach()
        max_rep = min_rep + torch.exp(delta_rep)
        boxes = Box(min_rep, max_rep)
        return boxes

    def transitive_rule_loss(self, ids):
        """
        Computes the loss for enforcing the transitive rule in the model.

        Args:
            ids: Tensor of triple ids.

        Returns:
            vol_loss: Volume loss enforcing the transitivity rule.
        """
        subsets = [ids[(ids[:,1] == r).nonzero().squeeze(1),:] for r in self.args.RULE_CONFIGS['transitive']['relations']]
        sub_ids = torch.cat(subsets, dim=0)

        # only optimize relation parameters
        head_boxes = self.get_entity_boxes_detached(sub_ids[:, 0])
        tail_boxes = self.get_entity_boxes_detached(sub_ids[:, 2])
        head_boxes = self.head_transformation(head_boxes, sub_ids[:,1])
        tail_boxes = self.tail_transformation(tail_boxes, sub_ids[:,1])

        intersection_boxes = self.intersection(head_boxes, tail_boxes)

        log_intersection = self.log_volumes(intersection_boxes)

        # P(f_r(epsilon_box)|g_r(epsilon_box)) should be 1
        vol_loss = torch.norm(1 - torch.exp(log_intersection - self.log_volumes(tail_boxes)))

        return vol_loss

    def composition_rule_loss(self, ids):
        """
        Computes the loss for enforcing the composition rule in the model.

        Args:
            ids: Tensor of triple ids.

        Returns:
            vol_loss: Volume loss enforcing the composition rule.
        """
        def rels(size, rid):
            # fill a tensor with relation id
            return torch.full((size,), rid, dtype=torch.long)

        def biconditioning(boxes1, boxes2):
            intersection_boxes = self.intersection(boxes1, boxes2)
            log_intersection = self.log_volumes(intersection_boxes)
            # || 1-P(Box2|Box1) ||
            condition_on_box1 = torch.norm(1 - torch.exp(log_intersection - self.log_volumes(boxes1)))
            # || 1-P(Box1|Box2) ||
            condition_on_box2 = torch.norm(1 - torch.exp(log_intersection - self.log_volumes(boxes2)))
            loss = condition_on_box1 + condition_on_box2

            return loss

        vol_loss = 0
        for rule_combn in self.args.RULE_CONFIGS['composite']['relations']:
            r1, r2, r3 = rule_combn
            r1_triples = ids[(ids[:, 1] == r1).nonzero().squeeze(1), :]
            r2_triples = ids[(ids[:, 1] == r2).nonzero().squeeze(1), :]

            # use heads and tails from r1, r2 as reasonable entity samples to help optimize relation parameters
            if len(r1_triples) > 0 and len(r2_triples) > 0:
                entities = torch.cartesian_prod(r1_triples[:,0], r2_triples[:,2])
                head_ids, tail_ids = entities[:,0], entities[:,1]
                size = len(entities)

                # only optimize relation parameters
                head_boxes_r1r2 = self.get_entity_boxes_detached(head_ids)
                tail_boxes_r1r2 = self.get_entity_boxes_detached(tail_ids)
                r1r2_head = self.head_transformation(head_boxes_r1r2, rels(size, r1))
                r1r2_head = self.head_transformation(r1r2_head, rels(size, r2))
                r1r2_tail = self.tail_transformation(tail_boxes_r1r2, rels(size, r1))
                r1r2_tail = self.tail_transformation(r1r2_tail, rels(size, r2))

                # head_boxes_r1r2 have been modified in transformation
                # so make separate box objects with the same parameters
                head_boxes_r3 = self.get_entity_boxes_detached(head_ids)
                tail_boxes_r3 = self.get_entity_boxes_detached(tail_ids)
                r3_head = self.head_transformation(head_boxes_r3, rels(size, r3))
                r3_tail = self.tail_transformation(tail_boxes_r3, rels(size, r3))

                head_transform_loss = biconditioning(r1r2_head, r3_head)
                tail_transform_loss = biconditioning(r1r2_tail, r3_tail)
                vol_loss += head_transform_loss
                vol_loss += tail_transform_loss
        return vol_loss


class Box:
    """
    A class representing an n-dimensional axis-aligned hyperrectangle (box) in the embedding space.

    Attributes:
        min_embed: The minimum boundary vector of the box in the embedding space.
        max_embed: The maximum boundary vector of the box in the embedding space.
        delta_embed: The difference between the maximum and minimum boundaries, representing the size of the box in each dimension.
    """
    def __init__(self, min_embed, max_embed):
        self.min_embed = min_embed
        self.max_embed = max_embed
        self.delta_embed = max_embed - min_embed



from .UKGE import UKGE
from .BEUrRE import BEUrRE
from .FocusE import FocusE
from .UKGE_PSL import UKGE_PSL
from .PASSLEAF import PASSLEAF
from .UKGsE import UKGsE
from .GTransE import GTransE
from .UPGAT import UPGAT